package sun.font;

import java.lang.ref.ReferenceQueue;
import sun.java2d.Disposer;
import sun.java2d.DisposerRecord;
import java.lang.ref.SoftReference;
import java.awt.font.TextLayout;
import java.io.ObjectOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.awt.GraphicsEnvironment;
import java.util.concurrent.ConcurrentHashMap;
import java.awt.font.FontRenderContext;
import java.awt.geom.AffineTransform;
import java.awt.Font;
import java.awt.FontMetrics;

public final class FontDesignMetrics extends FontMetrics
{
    static final long serialVersionUID = 4480069578560887773L;
    private static final float UNKNOWN_WIDTH = -1.0f;
    private static final int CURRENT_VERSION = 1;
    private static float roundingUpValue;
    private Font font;
    private float ascent;
    private float descent;
    private float leading;
    private float maxAdvance;
    private double[] matrix;
    private int[] cache;
    private int serVersion;
    private boolean isAntiAliased;
    private boolean usesFractionalMetrics;
    private AffineTransform frcTx;
    private transient float[] advCache;
    private transient int height;
    private transient FontRenderContext frc;
    private transient double[] devmatrix;
    private transient FontStrike fontStrike;
    private static FontRenderContext DEFAULT_FRC;
    private static final ConcurrentHashMap<Object, FontDesignMetrics.KeyReference> metricsCache;
    private static final int MAXRECENT = 5;
    private static final FontDesignMetrics[] recentMetrics;
    private static int recentIndex;
    
    private static FontRenderContext getDefaultFrc() {
        if (FontDesignMetrics.DEFAULT_FRC == null) {
            AffineTransform defaultTransform;
            if (GraphicsEnvironment.isHeadless()) {
                defaultTransform = new AffineTransform();
            }
            else {
                defaultTransform = GraphicsEnvironment.getLocalGraphicsEnvironment().getDefaultScreenDevice().getDefaultConfiguration().getDefaultTransform();
            }
            FontDesignMetrics.DEFAULT_FRC = new FontRenderContext(defaultTransform, false, false);
        }
        return FontDesignMetrics.DEFAULT_FRC;
    }
    
    public static FontDesignMetrics getMetrics(Font font) {
        return getMetrics(font, getDefaultFrc());
    }
    
    public static FontDesignMetrics getMetrics(Font font, FontRenderContext fontRenderContext) {
        if (SunFontManager.getInstance().maybeUsingAlternateCompositeFonts() && FontUtilities.getFont2D(font) instanceof CompositeFont) {
            return new FontDesignMetrics(font, fontRenderContext);
        }
        Object o = null;
        int equals = fontRenderContext.equals(getDefaultFrc());
        FontDesignMetrics.KeyReference keyReference;
        if (!(equals == 0)) {
            keyReference = (FontDesignMetrics.KeyReference)FontDesignMetrics.metricsCache.get(font);
        }
        else {
            synchronized (FontDesignMetrics.MetricsKey.class) {
                FontDesignMetrics.MetricsKey.key.init(font, fontRenderContext);
                keyReference = (FontDesignMetrics.KeyReference)FontDesignMetrics.metricsCache.get(FontDesignMetrics.MetricsKey.key);
            }
        }
        if (keyReference != null) {
            o = (FontDesignMetrics)keyReference.get();
        }
        if (o == null) {
            o = new FontDesignMetrics(font, fontRenderContext);
            if (!(equals == 0)) {
                FontDesignMetrics.metricsCache.put(font, new KeyReference(font, o));
            }
            else {
                FontDesignMetrics.MetricsKey key = new FontDesignMetrics.MetricsKey(font, fontRenderContext);
                FontDesignMetrics.metricsCache.put(key, new KeyReference(key, o));
            }
        }
        int i = 0;
        while (i < FontDesignMetrics.recentMetrics.length) {
            if (FontDesignMetrics.recentMetrics[i] == o) {
                return o;
            }
            ++i;
        }
        synchronized (FontDesignMetrics.recentMetrics) {
            FontDesignMetrics.recentMetrics[FontDesignMetrics.recentIndex++] = o;
            if (FontDesignMetrics.recentIndex == 5) {
                FontDesignMetrics.recentIndex = 0;
            }
        }
        return o;
    }
    
    private FontDesignMetrics(Font font) {
        this(font, getDefaultFrc());
    }
    
    private FontDesignMetrics(Font font, FontRenderContext frc) {
        super(font);
        this.serVersion = 0;
        this.height = -1;
        this.devmatrix = null;
        this.font = font;
        this.frc = frc;
        this.isAntiAliased = frc.isAntiAliased();
        this.usesFractionalMetrics = frc.usesFractionalMetrics();
        this.frcTx = frc.getTransform();
        this.matrix = new double[4];
        this.initMatrixAndMetrics();
        this.initAdvCache();
    }
    
    private void initMatrixAndMetrics() {
        this.fontStrike = FontUtilities.getFont2D(this.font).getStrike(this.font, this.frc);
        StrikeMetrics fontMetrics = this.fontStrike.getFontMetrics();
        this.ascent = fontMetrics.getAscent();
        this.descent = fontMetrics.getDescent();
        this.leading = fontMetrics.getLeading();
        this.maxAdvance = fontMetrics.getMaxAdvance();
        this.devmatrix = new double[4];
        this.frcTx.getMatrix(this.devmatrix);
    }
    
    private void initAdvCache() {
        this.advCache = new float[256];
        for (int i = 0; i < 256; ++i) {
            this.advCache[i] = -1.0f;
        }
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws IOException, ClassNotFoundException {
        objectInputStream.defaultReadObject();
        if (this.serVersion != 1) {
            this.frc = getDefaultFrc();
            this.isAntiAliased = this.frc.isAntiAliased();
            this.usesFractionalMetrics = this.frc.usesFractionalMetrics();
            this.frcTx = this.frc.getTransform();
        }
        else {
            this.frc = new FontRenderContext(this.frcTx, this.isAntiAliased, this.usesFractionalMetrics);
        }
        this.height = -1;
        this.cache = null;
        this.initMatrixAndMetrics();
        this.initAdvCache();
    }
    
    private void writeObject(ObjectOutputStream objectOutputStream) throws IOException {
        this.cache = new int[256];
        for (int i = 0; i < 256; ++i) {
            this.cache[i] = -1;
        }
        this.serVersion = 1;
        objectOutputStream.defaultWriteObject();
        this.cache = null;
    }
    
    private float handleCharWidth(int n) {
        return this.fontStrike.getCodePointAdvance(n);
    }
    
    private float getLatinCharWidth(char c) {
        float handleCharWidth = this.advCache[c];
        if (handleCharWidth == -1.0f) {
            handleCharWidth = this.handleCharWidth((int)c);
            this.advCache[c] = handleCharWidth;
        }
        return handleCharWidth;
    }
    
    @Override
    public FontRenderContext getFontRenderContext() {
        return this.frc;
    }
    
    @Override
    public int charWidth(char c) {
        float n;
        if (c < 'Ā') {
            n = this.getLatinCharWidth(c);
        }
        else {
            n = this.handleCharWidth((int)c);
        }
        return (int)(0.5 + (double)n);
    }
    
    @Override
    public int charWidth(int codePoint) {
        if (!Character.isValidCodePoint(codePoint)) {
            codePoint = 65535;
        }
        return (int)(0.5 + (double)this.handleCharWidth(codePoint));
    }
    
    @Override
    public int stringWidth(String s) {
        float n = 0.0f;
        if (this.font.hasLayoutAttributes()) {
            if (s == null) {
                throw new NullPointerException("str is null");
            }
            if (s.length() == 0) {
                return 0;
            }
            n = new TextLayout(s, this.font, this.frc).getAdvance();
        }
        else {
            for (int length = s.length(), i = 0; i < length; ++i) {
                int char1 = s.charAt(i);
                if (char1 < 'Ā') {
                    n += this.getLatinCharWidth((char)char1);
                }
                else {
                    if (FontUtilities.isNonSimpleChar((char)char1)) {
                        n = new TextLayout(s, this.font, this.frc).getAdvance();
                        break;
                    }
                    n += this.handleCharWidth(char1);
                }
            }
        }
        return (int)(0.5 + (double)n);
    }
    
    @Override
    public int charsWidth(char[] array, int i, int count) {
        float n = 0.0f;
        if (this.font.hasLayoutAttributes()) {
            if (count == 0) {
                return 0;
            }
            n = new TextLayout(new String(array, i, count), this.font, this.frc).getAdvance();
        }
        else {
            if (count < 0) {
                throw new IndexOutOfBoundsException("len=" + count);
            }
            int n2;
            int n3;
            for (n2 = i + count, ; i < n2; ++i) {
                n3 = array[i];
                if (n3 < 'Ā') {
                    n += this.getLatinCharWidth((char)n3);
                }
                else {
                    if (FontUtilities.isNonSimpleChar((char)n3)) {
                        n = new TextLayout(new String(array, i, count), this.font, this.frc).getAdvance();
                        break;
                    }
                    n += this.handleCharWidth(n3);
                }
            }
        }
        return (int)(0.5 + (double)n);
    }
    
    @Override
    public int[] getWidths() {
        int[] array = new int[256];
        float n;
        for (int i = 0; i < 256; array[i] = (int)(0.5 + (double)n), i = (char)(i + 1)) {
            n = this.advCache[i];
            if (n == -1.0f) {
                float[] advCache = this.advCache;
                int n2 = i;
                float handleCharWidth = this.handleCharWidth(i);
                advCache[n2] = handleCharWidth;
                n = handleCharWidth;
            }
        }
        return array;
    }
    
    @Override
    public int getMaxAdvance() {
        return (int)(0.99f + this.maxAdvance);
    }
    
    @Override
    public int getAscent() {
        return (int)(FontDesignMetrics.roundingUpValue + this.ascent);
    }
    
    @Override
    public int getDescent() {
        return (int)(FontDesignMetrics.roundingUpValue + this.descent);
    }
    
    @Override
    public int getLeading() {
        return (int)(FontDesignMetrics.roundingUpValue + this.descent + this.leading) - (int)(FontDesignMetrics.roundingUpValue + this.descent);
    }
    
    @Override
    public int getHeight() {
        if (this.height < 0) {
            this.height = this.getAscent() + (int)(FontDesignMetrics.roundingUpValue + this.descent + this.leading);
        }
        return this.height;
    }
    
    static {
        FontDesignMetrics.roundingUpValue = 0.95f;
        FontDesignMetrics.DEFAULT_FRC = null;
        metricsCache = new ConcurrentHashMap();
        recentMetrics = new FontDesignMetrics[5];
        FontDesignMetrics.recentIndex = 0;
    }
    
    private static class KeyReference extends SoftReference implements DisposerRecord, Disposer.PollDisposable
    {
        static ReferenceQueue queue;
        Object key;
        
        KeyReference(Object key, Object referent) {
            super(referent, FontDesignMetrics.KeyReference.queue);
            this.key = key;
            Disposer.addReference(this, this);
        }
        
        @Override
        public void dispose() {
            if (FontDesignMetrics.metricsCache.get(this.key) == this) {
                FontDesignMetrics.metricsCache.remove(this.key);
            }
        }
        
        static {
            FontDesignMetrics.KeyReference.queue = Disposer.getQueue();
        }
    }
    
    private static class MetricsKey
    {
        Font font;
        FontRenderContext frc;
        int hash;
        static final FontDesignMetrics.MetricsKey key;
        
        MetricsKey() {
        }
        
        MetricsKey(Font font, FontRenderContext fontRenderContext) {
            this.init(font, fontRenderContext);
        }
        
        void init(Font font, FontRenderContext frc) {
            this.font = font;
            this.frc = frc;
            this.hash = font.hashCode() + frc.hashCode();
        }
        
        @Override
        public boolean equals(Object o) {
            return o instanceof FontDesignMetrics.MetricsKey && this.font.equals(((FontDesignMetrics.MetricsKey)o).font) && this.frc.equals(((FontDesignMetrics.MetricsKey)o).frc);
        }
        
        @Override
        public int hashCode() {
            return this.hash;
        }
        
        static {
            key = new FontDesignMetrics.MetricsKey();
        }
    }
}
